(ns puppetlabs.activity.postgres
  (:import java.sql.Timestamp)
  (:require [clojure.java.jdbc :as jdbc]
            [migratus.core :as migratus]
            [clj-time.core :as time]
            [clj-time.format :as fmt-time]
            [puppetlabs.kitchensink.core :as ks]
            [slingshot.slingshot :refer [throw+ try+]]))

(def default-event-commits-limit 1000)
(def default-flat-event-commits-limit 10000)

;; TODO once puppetlabs.jdbc-util.core is available in nexus remove 'convert-result-arrays' & 'query' functions
;; TODO from this file and use the 'query' function from jdbc-util
(defn- convert-result-arrays
  "Converts Java and JDBC arrays in a result set using the provided function
   (eg. vec, set). Values which aren't arrays are unchanged."
  ([result-set]
   (convert-result-arrays vec result-set))
  ([f result-set]
   (let [convert #(cond
                   (ks/array? %) (f %)
                   (isa? (class %) java.sql.Array) (f (.getArray %))
                   :else %)]
     (map #(ks/mapvals convert %) result-set))))

(defn- query
  "An implementation of query that returns a fully evaluated result (no
   JDBCArray objects, etc)"
  [db sql-and-params]
  (let [convert (fn [rs]
                  (doall
                    (convert-result-arrays vec
                                           (jdbc/result-set-seq rs))))]
    (jdbc/db-query-with-resultset db sql-and-params convert)))
;; end of TODO

(defn public-tables
  "Get the names of all public tables in a database"
  [db-spec]
  (let [selection "SELECT table_name FROM information_schema.tables WHERE LOWER(table_schema) = 'public'"
        results (query db-spec [selection])]
    (map :table_name results)))

(defn drop-public-tables!
  "Drops all public tables in a database. Super dangerous."
  [db-spec]
  (when-let [tables (seq (public-tables db-spec))]
    (apply jdbc/db-do-commands db-spec (map #(format "DROP TABLE %s CASCADE" %) (seq tables)))))

(defn public-functions
  "Get the names of all public functions in a database"
  [db-spec]
  (let [selection (str "SELECT ns.nspname || '.' || proname || '(' || oidvectortypes(proargtypes) || ')' AS function"
                       "  FROM pg_proc INNER JOIN pg_namespace ns ON (pg_proc.pronamespace = ns.oid)"
                       "  WHERE ns.nspname = 'public'")
        results (query db-spec [selection])]
    (map :function results)))

(defn drop-public-functions!
  "Drops all public functions in a database. Super dangerous."
  [db-spec]
  (when-let [functions (seq (public-functions db-spec))]
    (apply jdbc/db-do-commands db-spec (map #(format "DROP FUNCTION %s CASCADE" %) (seq functions)))))

(defn migrate! [db]
  (migratus/migrate {:store         :database
                     :migration-dir "puppetlabs/activity/migrations"
                     :db            db}))

(defrecord Postgres [db])

(defn new-db [spec]
  (Postgres. spec))

(def event-commits-selection
  "SELECT ec.id                 AS id,
          sd.subject_type       AS subject_type,
          sd.subject_id         AS subject_id,
          sd.subject_name       AS subject_name,
          od.object_type        AS object_type,
          od.object_id          AS object_id,
          od.object_name        AS object_name,
          ec.submit_time        AS submit_time
  FROM event_commits ec
  JOIN subject_digests AS sd ON sd.id = ec.subject_digest_id
  JOIN object_digests AS od ON od.id = ec.object_digest_id")

(def event-commits-count-selection
  "SELECT count(*) AS count
  FROM event_commits ec
  JOIN subject_digests AS sd ON sd.id = ec.subject_digest_id
  JOIN object_digests AS od ON od.id = ec.object_digest_id")

(def flat-events-selection
  "SELECT ec.submit_time,
          sd.subject_type,
          sd.subject_id,
          sd.subject_name,
          od.object_type,
          od.object_id,
          od.object_name,
          ev.action_type,
          ev.action_what,
          ev.action_description,
          ev.message
  FROM event_commits AS ec
  JOIN subject_digests AS sd ON sd.id = ec.subject_digest_id
  JOIN object_digests AS od ON od.id = ec.object_digest_id
  JOIN events AS ev ON ev.commit_id = ec.id")

(defn- question-marks
  [count]
  (clojure.string/join ", " (repeat count "?")))

(defn- smart-split
  "The same functionality as clojure.string/split, however, this one handles escaped delimiters"
  [string delim-char]
  (let [re (re-pattern (str "^((?:[^\\\\" delim-char "]|\\\\.)+)(?:" delim-char "(.*))?$"))]
    (loop [s string result []]
      (if-let [groups (re-find re s)]
        (let [result (conj result (nth groups 1))]
          (if-let [remaining (nth groups 2)]
            (recur remaining result)
            result))
        result))))

(defn- csl->vector
  "Converts a string that contains multiple items separated by commas into a vector of items - e.g. 'kb,fh' to ['kb' 'fh']"
  [items]
  (if (nil? items)
    []
    (smart-split items ",")))

(def empty-item ["" []])

(defn- event-commits-condition-and-params
  "Concatenates individual segments of a where condition and at the same time collects parameters of the condition.
   Example of the final outcome: [\" WHERE ec.service_id = ? AND  sd.subject_type = ?\" [\"classifier\" \"users\"]]"
  [{:keys [service_id subject_type subject_id object_type object_id]}]
  (reduce (fn[[sql-segment1 params1] [sql-segment2 params2]]
            (vector (str sql-segment1 sql-segment2) (into params1 params2)))
          [[" WHERE ec.service_id = ?" [service_id]]
           (if subject_type
             [" AND sd.subject_type = ?" [subject_type]]
             empty-item)
           (if (not (empty? subject_id))
             (let [subject-ids (csl->vector subject_id)]
               [(str " AND sd.subject_id in (" (question-marks (count subject-ids)) ")") subject-ids])
             empty-item)
           (if object_type
             [" AND od.object_type = ?" [object_type]]
             empty-item)
           (if (not (empty? object_id))
             (let [object-ids (csl->vector object_id)]
               [(str " AND od.object_id in (" (question-marks (count object-ids)) ")") object-ids])
             empty-item)]))

(defn- order-by-submit-time
  [selection]
  (str selection " ORDER BY ec.submit_time DESC"))

(defn- order-by-submit-time-and-action-type-and-what
  [selection]
  (str selection " ORDER BY ec.submit_time DESC, ev.action_type, ev.action_what"))

(defn- select-event-commits
  [params]
  (let [condition-and-params (event-commits-condition-and-params params)]
    (into [(order-by-submit-time (str event-commits-selection (first condition-and-params)))]
          (second condition-and-params))))

(defn- select-event-commits-count
  [params]
  (let [condition-and-params (event-commits-condition-and-params params)]
    (into [(str event-commits-count-selection (first condition-and-params))]
          (second condition-and-params))))

(defn- select-flat-events
  [params]
  (let [condition-and-params (event-commits-condition-and-params params)]
    (into [(order-by-submit-time-and-action-type-and-what (str flat-events-selection (first condition-and-params)))]
          (second condition-and-params))))

(defn- add-offset-or-limit-item
  [sql-and-params item-sql item-value]

  (if item-value
    (let [sql (str (first sql-and-params) item-sql)]
      (conj (assoc sql-and-params 0 sql) (Integer. item-value)))
    sql-and-params))

(defn- add-offset-and-limit
  [sql-and-params offset limit]

  (let [with-offset (add-offset-or-limit-item sql-and-params " OFFSET ?" offset)]
    (add-offset-or-limit-item with-offset " LIMIT ?" limit)))

(def events-by-commit-selection
  "SELECT id                 AS id,
          action_type        AS action_type,
          action_what        AS action_what,
          action_description AS action_description,
          message            AS message
    FROM events
    WHERE commit_id = ?
    ORDER BY message")

(defn- select-events-by-commit
  [commit-id]
  [events-by-commit-selection commit-id])

(defn- transform-event-for-output
  [{:keys [message] :as row}]
  {:message message})

(defn- get-commit-with-events
  [db commit-id event-commit]
  (assoc event-commit :events (mapv #(transform-event-for-output %)
                                    (query db (select-events-by-commit commit-id)))))

(defn- transform-event-commit-for-output
  [{:keys [subject_name object_id submit_time]}]
  {:object    object_id
   :subject   subject_name
   :timestamp submit_time})

(defn check-rbac-permissions
  "Checks RBAC parmissions. Throws an exception if the permissions are not sufficient."
  [is-current-subject-permitted? permission-string]
  (when (not (is-current-subject-permitted? permission-string))
    (throw+ {:kind :puppetlabs.rbac/permission-denied :msg "Insufficient permissions to get activity events."})))

(defn check-full-permissions
  [is-current-subject-permitted?]
  (check-rbac-permissions is-current-subject-permitted? "activity_log:view"))

(defn check-object-permissions
  [is-current-subject-permitted? object_type object_id]
  (check-rbac-permissions is-current-subject-permitted? (format "%s:view:%s" object_type object_id)))

(defn check-permissions
  [is-current-subject-permitted? {:keys [object_type object_id]}]
  (try+
    (check-full-permissions is-current-subject-permitted?)
    (catch [:kind :puppetlabs.rbac/permission-denied] map
      (if (and object_type object_id)
        (check-object-permissions is-current-subject-permitted? object_type object_id)
        (throw+ map)))))

(defn- is-number?
  "Checks if the string parameter contains a number."
  [str]
  (number? (ks/parse-number str)))

(defn- params-ok?
  [{:keys [service_id subject_type subject_id object_type object_id offset limit]}]

  (and service_id
       (or (and subject_type subject_id) (not subject_id))
       (or (and object_type object_id) (not object_id))
       (or (nil? offset) (is-number? offset))
       (or (nil? limit) (is-number? limit))))

(defn get-events
  [{db :db} is-current-subject-permitted? {:keys [offset limit] :as params}]

  {:pre [(params-ok? params)]}

  (check-permissions is-current-subject-permitted? params)

  (let [safe-limit (or limit default-event-commits-limit)
        commits (mapv
                  #(get-commit-with-events db (:id %) (transform-event-commit-for-output %))
                  (let [sql-and-params (add-offset-and-limit (select-event-commits params) offset safe-limit)]
                    (query db sql-and-params)))
        total-rows (:count
                    (first (let [sql-and-params (select-event-commits-count params)]
                             (query db sql-and-params))))]
    {:commits    commits
     :offset     (or offset 0)
     :limit      (or limit total-rows)
     :total-rows total-rows}))

(def column-names-map
  {:submit_time        "Submit Time"
   :subject_type       "Subject Type"
   :subject_id         "Subject Id"
   :subject_name       "Subject Name"
   :object_type        "Object Type"
   :object_id          "Object Id"
   :object_name        "Object Name"
   :action_type        "Type"
   :action_what        "What"
   :action_description "Description"
   :message            "Message"})

(defn get-flat-events
  [{db :db} is-current-subject-permitted? {:keys [offset limit] :as params}]

  {:pre [(params-ok? params)]}

  (check-permissions is-current-subject-permitted? params)

  (let [safe-limit (or limit default-flat-event-commits-limit)
        sql-and-params (add-offset-and-limit (select-flat-events params) offset safe-limit)
        result-set (jdbc/query db sql-and-params :as-arrays? true)]

    (assoc result-set
      0
      (mapv #(if-let [name (column-names-map %)] name %)
            (get result-set 0)))))

(defn- parse-iso8601-datetime
  [datetime]
  (Timestamp.
    (.getMillis
      (let [iso8601-formatter (fmt-time/formatters :date-time-no-ms)]
        (fmt-time/parse iso8601-formatter datetime)))))

(defn- current-time
  []
  (Timestamp. (.getMillis (time/now))))

(defn- unique-savepoint-name
  "Generates a unique savepoint name."
  []
  (str "sp_" (ks/utf8-string->sha1 (ks/uuid))))

(defn- insert-or-keep-original
  "Tries to do 'jdbc/insert!'. If a unique key constraint is violeted then it keeps the original record."
  [db-spec table row-map]

  (let [safepoint-name (unique-savepoint-name)]
    (try
      (jdbc/execute! db-spec [(str "SAVEPOINT " safepoint-name ";")])
      (jdbc/insert! db-spec table row-map)
      (jdbc/execute! db-spec [(str "RELEASE SAVEPOINT " safepoint-name ";")])
      (catch org.postgresql.util.PSQLException e
        (if (= (.getSQLState e) "23505")
          (jdbc/execute! db-spec [(str "ROLLBACK TO SAVEPOINT " safepoint-name ";")])
          (throw e))))))

(defn- type-id-name-hash
  [map]
  (ks/utf8-string->sha1 (str (:type map) (:id map) (:name map))))

;; aliases
(def subject-hash type-id-name-hash)
(def object-hash type-id-name-hash)

(defn- commit-hash
  [service-id subject-digest-id object-digest-id submit-time]
  (ks/utf8-string->sha1 (str service-id subject-digest-id object-digest-id submit-time)))

(defn- event-hash
  [commit-id action-type action-what action-description]
  (ks/utf8-string->sha1 (str commit-id action-type action-what action-description)))

(defn store-events!
  [{db :db} data]

  (let [commit (:commit data)
        {:keys [service subject object events]} commit]

    (jdbc/with-db-transaction
      [t-db db]

      (let [subject-digest-id (subject-hash subject)
            object-digest-id (object-hash object)
            submit-time (current-time)
            commit-id (commit-hash (:id service) subject-digest-id object-digest-id submit-time)]

        (insert-or-keep-original
          t-db
          :subject_digests
          {:id           subject-digest-id
           :subject_type (:type subject)
           :subject_id   (:id subject)
           :subject_name (:name subject)})

        (insert-or-keep-original
          t-db
          :object_digests
          {:id          object-digest-id
           :object_type (:type object)
           :object_id   (:id object)
           :object_name (:name object)})

        (jdbc/insert!
          t-db
          :event_commits
          {:id                commit-id
           :service_id        (:id service)
           :subject_digest_id subject-digest-id
           :object_digest_id  object-digest-id
           :submit_time       submit-time})
        (mapv #(jdbc/insert!
                t-db
                :events
                {:id                 (event-hash commit-id (:type %) (:what %) (:description %))
                 :commit_id          commit-id
                 :action_type        (:type %)
                 :action_what        (:what %)
                 :action_description (:description %)
                 :message            (:message %)})
              events)))))
